import { world, system } from "@minecraft/server";
import { ActionFormData, MessageFormData, ModalFormData } from "@minecraft/server-ui";
import config from "./config.js";
import {teleportPlayerToSpawn, teleportPlayerToPlayer, teleportPlayerToDXYZ, teleportPlayerToDXYZFacingXYZ, teleportPlayerToHome, setHomeTeleport, hasHomeTeleport, getPlayerWaypoints, setPlayerWaypoint, teleportPlayerToWaypoint, getServerWaypoints, removePlayerWaypoint, setServerWaypoint, removeServerWaypoint} from './util/teleport.js';
import {showActionFormButtonList, showAreYouSureMenu} from './ui/ui_helper.js';
import translate from "./util/translation_helper.js";

const ICONS = {
    home: "textures/ui_icons/house_128",
    spawn_teleport: "textures/ui_icons/map_waypoint_128",
    server_waypoint: "textures/ui_icons/waypoint_blue_128",
    server_waypoint_menu: "textures/ui_icons/waypoint_double_blue_128",
    player_waypoint: "textures/ui_icons/waypoint_128",
    player_waypoint_menu: "textures/ui_icons/waypoint_double_128",
    player: "textures/ui_icons/player_128",
    player_teleport_menu: "textures/ui_icons/player_to_player_path_128",
    check: "textures/ui_icons/check_128",
    x: "textures/ui_icons/x_128",
    gear: "textures/ui_icons/gear_128",
    toggle_on: "textures/ui_icons/toggle_on_128",
    toggle_off: "textures/ui_icons/toggle_off_128"
}

if (config.enableSpawnTeleport) {
    let spawnLocation = world.getDefaultSpawnLocation();
    world.getDimension("minecraft:overworld").runCommand(`tickingarea add circle ${spawnLocation.x} ${spawnLocation.y} ${spawnLocation.z} 1 tb:spawn true`);    
} else {
    world.getDimension("minecraft:overworld").runCommand(`tickingarea remove tb:spawn`);
}

world.afterEvents.playerSpawn.subscribe((event) => {
    let player = event.player;

    if (config.giveCompassOnPlayerSpawn) {
        player.runCommandAsync('execute as @s if entity @s [tag=!joined] run give @s compass');
    }
});

world.afterEvents.itemUse.subscribe(event => {
    const player = event.source;

    if (event.itemStack.typeId == "minecraft:compass") {
        if (player.isSneaking) {
            if (true) {
                if (hasHomeTeleport(player)) {
                    showSetHomeConfirmationMenu(player);
                } else {
                    setHomeTeleport(player, player.dimensionId, player.location.x, player.location.y, player.location.z);
                }                
            } else { // TODO: Add limitations to where you can set home?
                //player.runCommandAsync('/titleraw @s actionbar {"rawtext":[{"text":" §l§cYou Can Not Set Your §dHome §cHere !!"}]}');
                //player.runCommandAsync("playsound note.bass @s");
            }
        } else if (!player.isSneaking) {
            showTeleportMenu(player);
        }

    }
});

function showSetHomeConfirmationMenu(player) {
    showAreYouSureMenu(player, translate("tbu.teleport.move_house_title"), translate("tbu.teleport.move_house_body"), () => setHomeTeleport(player, player.dimensionId, player.location.x, player.location.y, player.location.z))
}


function showTeleportMenu(player) {
    let fm = new ActionFormData();
    fm.title(translate("tbu.teleport.main_menu_title"));

    let index = -1;
    let callbacks = [];

    // Home
    if (config.enableHomes || player.hasTag("op")) {
        if (hasHomeTeleport(player)) {
            fm.button(translate("tbu.teleport.home"), ICONS.home);
            callbacks[++index] = () => {teleportPlayerToHome(player)};
        } else {
            fm.button(translate("tbu.teleport.set_home"), ICONS.home);
            callbacks[++index] = () => {setHomeTeleport(player)};
        }
        
    }

    // Server Waypoints
    if (config.enableServerWaypoints || player.hasTag("op")) {
        if (config.enableSpawnTeleport || player.hasTag("op")) {
            fm.button(translate("tbu.teleport.spawn"), ICONS.spawn_teleport);
            callbacks[++index] = () => {teleportPlayerToSpawn(player)};
        }

        fm.button(translate("tbu.teleport.server_waypoints"), ICONS.server_waypoint_menu);
        callbacks[++index] = () => {showServerWaypointMenu(player)};
    }

    // Player Waypoints
    if (config.enablePlayerWaypoints || player.hasTag("op")) {
        fm.button(translate("tbu.teleport.player_waypoints"), ICONS.player_waypoint_menu);
        callbacks[++index] = () => {showPlayerWaypointMenu(player)};
    }

    // Player to Player Teleports
    if (config.enablePlayerToPlayerTeleports || player.hasTag("op")) {
        fm.button(translate("tbu.teleport.players"), ICONS.player_teleport_menu);
        callbacks[++index] = () => {showPlayerTeleportMenu(player)};
    }

    // Settings
    fm.button(translate("tbu.teleport.settings"), ICONS.gear);
    callbacks[++index] = () => {showSettingsMenu(player)};

    // Show form
    fm.show(player).then(result => {
        if (result.selection < callbacks.length) {
            callbacks[result.selection]();
        }
    });
}

function showServerWaypointMenu(player) {
    let buttons = [];
    if (config.enableSpawnTeleport || player.hasTag("op")) {
        buttons.push({
            name: translate("tbu.teleport.spawn"),
            icon: ICONS.spawn_teleport,
            callback: () => {teleportPlayerToSpawn(player)}
        });
    }

    let serverWaypoints = getServerWaypoints(player);

    serverWaypoints.forEach((wp) => {
        buttons.push({
            name: wp.name,
            icon: ICONS.server_waypoint,
            callback: () => {teleportPlayerToWaypoint(player, wp)}
        })
    });

    if (player.hasTag("op")) {
        buttons.push({
            name: translate("tbu.teleport.set_server_waypoint"),
            callback: () => {showSetServerWaypointMenu(player)}
        });
        if (serverWaypoints.length > 0) {
            buttons.push({
                name: translate("tbu.teleport.remove_server_waypoint"),
                callback: () => {showRemoveServerWaypointMenu(player)}
            });
        }
    }

    if (buttons.length > 0) {
        showActionFormButtonList(player, translate("tbu.teleport.server_waypoints"), undefined, buttons);
    } else {
        showActionFormButtonList(player, translate("tbu.teleport.server_waypoints"), translate("tbu.teleport.no_waypoints_available"), [], true);
    }
    
}

function showPlayerWaypointMenu(player) {
    let playerWaypoints = getPlayerWaypoints(player);
    let buttons = [];
    playerWaypoints.forEach((wp) => {
        buttons.push({
            name: wp.name,
            icon: ICONS.player_waypoint,
            callback: () => {teleportPlayerToWaypoint(player, wp)}
        })
    });

    if (playerWaypoints.length < config.maxWaypointsPerPlayer) {
        buttons.push({
            name: translate("tbu.teleport.set_waypoint"),
            callback: () => {showSetWaypointMenu(player)}
        });
    }

    if (playerWaypoints.length > 0) {
        buttons.push({
            name: translate("tbu.teleport.remove_waypoint"),
            callback: () => {showRemoveWaypointMenu(player)}
        });
    }

    if (buttons.length > 0) {
        showActionFormButtonList(player, translate("tbu.teleport.player_waypoints"), translate("tbu.teleport.number_waypoints_established", [playerWaypoints.length, config.maxWaypointsPerPlayer]), buttons);
    } else {
        showActionFormButtonList(player, translate("tbu.teleport.player_waypoints"), translate("tbu.teleport.no_waypoints_available"), [], true);
    }
}

function showPlayerTeleportMenu(player) {
    let players = world.getPlayers().filter((p) => {
        if (p.id != player.id) {
            if (player.hasTag("op"))
                return true;
            else if (!config.allowHiddenPlayers || !p.hasTag("novisitors"))
                return true;
        }
            
        return false;
    });


    let buttons = [];
    players.forEach((targetPlayer) => {
        buttons.push({
            name: targetPlayer.nameTag,
            icon: ICONS.player,
            callback: () => {teleportPlayerToPlayer(player, targetPlayer)}
        });
    });

    if (buttons.length > 0) {
        showActionFormButtonList(player, translate("tbu.teleport.players"), undefined, buttons);
    } else {
        showActionFormButtonList(player, translate("tbu.teleport.players"), translate("tbu.teleport.no_players_available"), [], true);
    }
    
}


function showSettingsMenu(player) {
    let fm = new ActionFormData();
    fm.title(translate("tbu.teleport.settings_menu_title"));
    
    let index = -1;
    let callbacks = [];

    if (player.hasTag("op") || config.allowHiddenPlayers) {
        fm.button(translate("tbu.teleport.accept_visitors"), player.hasTag("novisitors") ? ICONS.toggle_off : ICONS.toggle_on);
        callbacks[++index] = () => {
            if (player.hasTag("novisitors")) {
                player.removeTag("novisitors");
            } else {
                player.addTag("novisitors");
            }
        };
    }

    if (player.hasTag("op") || config.enableHomes) {
        if (hasHomeTeleport(player)) {
            fm.button(translate("tbu.teleport.move_house"));
            callbacks[++index] = () => {showSetHomeConfirmationMenu(player);};
        } else {
            fm.button(translate("tbu.teleport.set_home"));
            callbacks[++index] = () => {setHomeTeleport(player)};
        }
    }

    if (player.hasTag("op") || config.enablePlayerWaypoints) {
        fm.button(translate("tbu.teleport.set_waypoint"));
        callbacks[++index] = () => {showSetWaypointMenu(player)};

        let playerWaypoints = getPlayerWaypoints(player);
        if (playerWaypoints.length > 0) {
            fm.button(translate("tbu.teleport.remove_waypoint"));
            callbacks[++index] = () => {showRemoveWaypointMenu(player)};
        }
    }

    if (player.hasTag("op") /*&& config.enableServerWaypoints*/) {
        fm.button(translate("tbu.teleport.set_server_waypoint"));
        callbacks[++index] = () => {showSetServerWaypointMenu(player)};

        let serverWaypoints = getServerWaypoints(player);
        if (serverWaypoints.length > 0) {
            fm.button(translate("tbu.teleport.remove_server_waypoint"));
            callbacks[++index] = () => {showRemoveServerWaypointMenu(player)};
        }
    }

    // TODO: Config menu

    fm.button(translate("tbu.teleport.cancel"));

    fm.show(player).then((result) => {
        if (result.selection < callbacks.length) {
            callbacks[result.selection]();
        }
    });
}

function showSetWaypointMenu(player) {
    // First confirm we are under the allowable waypoint limit. If not, offer to remove one
    let playerWaypoints = getPlayerWaypoints(player);
    if (playerWaypoints.length >= config.maxWaypointsPerPlayer) {
        // TODO force delete a waypoint
        return;
    }

    // Request a name for this waypoint
    let fm = new ModalFormData();
    fm.title(translate("tbu.teleport.set_waypoint"));
    fm.textField(translate("tbu.teleport.name_input"), translate("tbu.teleport.name_input_hint"), "Waypoint");
    if (config.allowPublicPlayerWaypoints || player.hasTag("op")) {
        fm.toggle(translate("tbu.teleport.make_public"), true);
    }
    
    fm.show(player).then((formData) => {
        let values = formData.formValues;
        // Add the waypoint to the list
        let name = values[0] == undefined || values[0] == "" ? "Waypoint" : values[0];
        let isPublic = false;
        if (config.allowPublicPlayerWaypoints || player.hasTag("op")) {
            isPublic = values[1] ? true : false;
        }
        
        setPlayerWaypoint(player, name, isPublic);
    });

}

function showRemoveWaypointMenu(player) {
    // First confirm we have a waypoint
    let playerWaypoints = getPlayerWaypoints(player);
    if (playerWaypoints.length == 0) {
        // TODO notify they have no waypoints
        return;
    }

    let buttons = [];
    playerWaypoints.forEach((wp) => {
        buttons.push({
            name: wp.name,
            icon: ICONS.player_waypoint,
            callback: () => {removePlayerWaypoint(player, wp)}
        })
    });

    showActionFormButtonList(player, translate("tbu.teleport.remove_waypoint"), {rawtext:[translate("tbu.teleport.number_waypoints_established", [playerWaypoints.length, config.maxWaypointsPerPlayer]), {text: "\n"}, translate("tbu.teleport.remove_which")]}, buttons, true);
}

function showSetServerWaypointMenu(player) {
    // Request a name for this waypoint
    let fm = new ModalFormData();
    fm.title(translate("tbu.teleport.set_server_waypoint"));
    fm.textField(translate("tbu.teleport.name_input"), translate("tbu.teleport.name_input_hint"), "Waypoint");

    fm.show(player).then((formData) => {
        let values = formData.formValues;

        let name = values[0] == undefined || values[0] == "" ? "Waypoint" : values[0];
        
        setServerWaypoint(player, name);
    });
}

function showRemoveServerWaypointMenu(player) {
    // First confirm we have a waypoint
    let serverWaypoints = getServerWaypoints(player);
    if (serverWaypoints.length == 0) {
        // TODO notify they have no waypoints
        return;
    }

    let buttons = [];
    serverWaypoints.forEach((wp) => {
        buttons.push({
            name: wp.name,
            icon: ICONS.server_waypoint,
            callback: () => {removeServerWaypoint(player, wp)}
        })
    });

    showActionFormButtonList(player, translate("tbu.teleport.remove_server_waypoint"), undefined, buttons, true);
}



